#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys, os
from bottle import route, run, response, request, static_file, HTTPError, auth_basic, ServerAdapter
from cherrypy import _cpwsgiserver3
import glob
import json
import time
import traceback
import ConfigParser
import logging
import logging.config
from datetime import datetime

restServiceHost = '0.0.0.0'
restServicePort = 28080
servicePath = os.path.dirname(os.path.abspath(__file__)) + '/'
callCount = {}
testMode = False
ruleList = {}
paymentList = {}
bindingList = {}
providerList = {}
organizationList = {}
terminalList = {}
organizationId = ""
paymentSystemList = {}

try:
    logging.config.fileConfig(servicePath + 'conf/logging.properties')
except Exception, e:
    print 'Ошибка `%s`' % e
    logging.basicConfig(format = u'%(levelname)-8s [%(asctime)s] %(message)s', level = logging.DEBUG, filename = u'paygatetest.log')

def getRule(funcName):
    rule = None
    func = funcName.lower()
    if testMode and func in ruleList:
        if func not in callCount:
            callCount.update({func: 0})
        callCount[func] += 1
        
        curNumber = callCount[func]
        if curNumber >= len(ruleList[func]) or ruleList[func][curNumber] == None: 
            curNumber = 0
        rule = ruleList[func][curNumber]
    return rule
def getResult(status,comment):
    return {u'status': status, u'comment': comment}

def checkPassword(username, password):
    global organizationId
    logging.debug(u'Аутентификация, login = %s' % username)
    for key,value in organizationList.items():
        if (username, password) == (value["login"], value["password"]):
            logging.debug(u'Организация = %s' % key)
            organizationId = key
            return True
    logging.error(u'Доступ запрещен')
    return False

def checkTerminal(cashCode, shopCode):
    logging.debug(u'Проверка терминала, касса = %s, магазин = %s, организация = %s'  % (cashCode, shopCode, organizationId))
    for key, value in terminalList.items():
        if(shopCode, cashCode, organizationId) ==(value["shopCode"], value["cashCode"], value["organization"]):
            logging.debug(u'Найден терминал %s' % key)
            return key
    logging.error(u'Для данного терминала невозможно выполнить операцию')
    return None

def checkPaymentSystem(provider, terminal):
    for key, value in paymentSystemList.items():
        if int(provider) in value["provider"] and int(terminal) in value["terminal"]:
            return key
    return None

def checkBaseRule(rule): 
    if rule == 'errorConnection':
        logging.error(u'Тест.Ошибка соединения')
        raise HTTPError(404)
    elif rule == 'authFailed':
        logging.error(u'Тест.Доступ запрещен')
        raise HTTPError(401)
    elif rule == 'terminalNotFound':
        logging.error(u'Тест.Терминал не найден')
        return 'Терминал не найден'
    return None

@route('/paygate3/api/bindings/cardnumber/<cardNum>', method='GET')
@auth_basic(checkPassword)
def getBindings(cardNum):
    cardNum = cardNum.decode('utf-8')
    logging.info(u'Получили запрос на получение списка привязок для карты %s' % cardNum)
    logging.info(u'URL: %s' % request.url)

    rule = getRule('getBindings')
    baseError = checkBaseRule(rule)
    if baseError:
        return json.dumps({u'result': getResult(300, baseError)}, ensure_ascii=False)
    if rule == 'cardNotFound':
        logging.error(u'Тест.Карта не найдена')
        return json.dumps({u'result': getResult(300, u'Карта не найдена')}, ensure_ascii=False)
    
    if cardNum not in bindingList:
        result = {u'result': getResult(300, u'Карта не найдена')}
        logging.error(u'Карта не найдена')
    else:
        bindings = bindingList[cardNum].values() if cardNum in bindingList else {}
        result = {u'result': getResult(0, u'OK'), u'bindings': bindings}
    logging.debug(u'Отправляем ответ: %s' % json.dumps(result, ensure_ascii=False))

    return json.dumps(result, ensure_ascii=False)

@route('/paygate3/api/providers/shopcode/<shopCode>/cashcode/<cashCode>/providercode/<providerCode>', method='GET')
@auth_basic(checkPassword)
def getProvider(shopCode, cashCode, providerCode):
    logging.info(u'Получили запрос на получение информации о поставщике')
    logging.info(u'URL: %s' % request.url)

    rule = getRule('getProvider')
    baseError = checkBaseRule(rule)
    if baseError:
        return json.dumps({u'result': getResult(300, baseError)}, ensure_ascii=False)

    if checkTerminal(cashCode, shopCode):
        result = {u'result': getResult(0, u'OK'), u'provider': providerList[providerCode]}
    else:
        result = {u'result': getResult(300, u'Для данного терминала невозможно выполнить операцию')}
    logging.debug(u'Отправляем ответ: %s' % json.dumps(result, ensure_ascii=False))   

    return json.dumps(result, ensure_ascii=False)

@route('/paygate3/api/providers/shopcode/<shopCode>/cashcode/<cashCode>', method='GET')
@auth_basic(checkPassword)
def getProviderList(shopCode, cashCode):
    logging.info(u'Получили запрос на получение списка поставщиков')
    logging.info(u'URL: %s' % request.url)

    rule = getRule('getProviderList')
    baseError = checkBaseRule(rule)
    if baseError:
        return json.dumps({u'result': getResult(300, baseError)}, ensure_ascii=False)

    if checkTerminal(cashCode, shopCode):
        result = {u'result': getResult(0, u'OK'), u'providers': providerList.values()}
    else:
        result = {u'result': getResult(300, u'Для данного терминала невозможно выполнить операцию')}
    logging.debug(u'Отправляем ответ: %s' % json.dumps(result, ensure_ascii=False))

    return json.dumps(result, ensure_ascii=False)

@route('/paygate3/api/payments/check', method='POST')
@auth_basic(checkPassword)
def checkPayment():
    logging.info(u'Получили запрос на проверку корректности платежа')
    logging.info(u'URL: %s' % request.url)
    body = json.loads(request.body.read(), encoding='utf8')
    logging.debug(u'Тело запроса: %s' % body)
    
    rule = getRule('checkPayment')
    baseError = checkBaseRule(rule)
    if baseError:
        return json.dumps({u'result': getResult(300, baseError)}, ensure_ascii=False)
    if rule == 'paymentSystemNotFound':
        logging.error(u'Тест.Платежная система не найдена')
        return json.dumps({u'result': getResult(300, u'Тест.Платежная система не найдена')}, ensure_ascii=False)

    terminalId = checkTerminal(request.query.get('cashCode'), request.query.get('shopCode'))
    if  terminalId == None:
        result = {u'result': getResult(300, u'Для данного терминала невозможно выполнить операцию')}
        logging.debug(u'Отправляем ответ: %s' % json.dumps(result, ensure_ascii=False))
        return json.dumps(result, ensure_ascii=False)

    paymentSystemId = checkPaymentSystem(body['payment']['providerCode'], terminalId)
    if  paymentSystemId == None:
        result = {u'result': getResult(300, u'Платежная система не найдена')}
        logging.error(u'Платежная система не найдена')
        logging.debug(u'Отправляем ответ: %s' % json.dumps(result, ensure_ascii=False))
        return json.dumps(result, ensure_ascii=False)

    paymentId = int(time.time()*1000)
    paymentList[paymentId] = body['payment']
    paymentList[paymentId]["paymentSystem"] = paymentSystemId
    result = {u'result': getResult(0, u'OK'), u'paymentId': paymentId}
    logging.debug(u'Отправляем ответ: %s' % json.dumps(result, ensure_ascii=False))

    return json.dumps(result, ensure_ascii=False)

@route('/paygate3/api/payments/checkresult/paymentid/<paymentId>', method='GET')
@auth_basic(checkPassword)
def checkPaymentResult(paymentId):
    logging.info(u'Получили запрос на результат проверки платежа')
    logging.info(u'URL: %s' % request.url)

    rule = getRule('checkPaymentResult')
    baseError = checkBaseRule(rule)
    if baseError:
        return json.dumps({u'result': getResult(300, baseError)}, ensure_ascii=False)
    if rule == 'paymentNotFound':
        logging.error(u'Тест.Платеж не найден')
        return json.dumps({u'result': getResult(300, u'Платеж не найден')}, ensure_ascii=False)
    elif rule == 'paymentDuring':
        logging.error(u'Тест.Платеж в обработке')
        return json.dumps({u'result': getResult(10, u'Платеж в обработке')}, ensure_ascii=False)


    if long(paymentId) not in paymentList:        
        result = {u'result': getResult(300, u'Платеж не найден')}
        logging.error(u'Платеж не найден')
        logging.debug(u'Отправляем ответ: %s' % json.dumps(result, ensure_ascii=False))
        return json.dumps(result, ensure_ascii=False)

    payment = paymentList[long(paymentId)]
    comissionRate = float(providerList[payment["providerCode"]]["comissionRate"])
    comission = float(payment["amount"]) * comissionRate / 100
    paymentResult = {
                    "paymentId":paymentId,
                    "date":datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f+0000'),
                    "comissionSum":comission,
                    "resultSum": float(payment["amount"]) - comission,
                    "contactInfo": {"organization" : organizationList[organizationId]["info"],
                                    "provider" : providerList[payment["providerCode"]]["info"],
                                    "paymentsystem": paymentSystemList[payment["paymentSystem"]]["info"]
                                   }
                    }
    result = {u'result': getResult(0, u'OK'), u'paymentResult': paymentResult}
    logging.debug(u'Отправляем ответ: %s' % json.dumps(result, ensure_ascii=False))

    return json.dumps(result, ensure_ascii=False)

def closePayment(paymentId):
    if checkTerminal(request.query.get('cashCode'), request.query.get('shopCode')) == None:
        result = {u'result': getResult(300, u'Для данного терминала невозможно выполнить операцию')}
    elif long(paymentId) not in paymentList:
        logging.error(u'Платеж не найден')
        result = {u'result': getResult(300, u'Платеж не найден')}
    else:
        result = {u'result': getResult(0, u'OK')}
    logging.debug(u'Отправляем ответ: %s' % json.dumps(result, ensure_ascii=False))

    return json.dumps(result, ensure_ascii=False)

@route('/paygate3/api/payments/pay/paymentid/<paymentId>', method='POST')
@auth_basic(checkPassword)
def commitPayment(paymentId):
    logging.info(u'Получили запрос на подтверждение платежа')
    logging.info(u'URL: %s' % request.url)

    rule = getRule('commitPayment')
    baseError = checkBaseRule(rule)
    if baseError:
        return json.dumps({u'result': getResult(300, baseError)}, ensure_ascii=False)
    if rule == 'paymentNotFound':
        logging.error(u'Тест.Платеж не найден')
        return json.dumps({u'result': getResult(300, u'Платеж не найден')}, ensure_ascii=False)

    return closePayment(paymentId)

@route('/paygate3/api/payments/cancel/paymentid/<paymentId>', method='PUT')
@auth_basic(checkPassword)
def cancelPayment(paymentId):
    logging.info(u'Получили запрос на отмену платежа')
    logging.info(u'URL: %s' % request.url)

    rule = getRule('cancelPayment')
    baseError = checkBaseRule(rule)
    if baseError:
        return json.dumps({u'result': getResult(300, baseError)}, ensure_ascii=False)
    if rule == 'paymentNotFound':
        logging.error(u'Тест.Платеж не найден')
        return json.dumps({u'result': getResult(300, u'Платеж не найден')}, ensure_ascii=False)

    return closePayment(paymentId)

@route('/paygate3/api/bindings/cardnumber/<cardNum>/accountnumber/<accNum>/providercode/<provCode>', method='POST')
@auth_basic(checkPassword)
def bindAccount(cardNum, accNum, provCode):
    cardNum = cardNum.decode('utf-8')
    logging.info(u'Получили запрос на привязку номера')
    logging.info(u'URL: %s' % request.url)

    rule = getRule('bindAccount')
    baseError = checkBaseRule(rule)
    if baseError:
        return json.dumps({u'result': getResult(300, baseError)}, ensure_ascii=False)
    if rule == 'accountLimit':
        logging.error(u'Тест.Превышен лимит привязок')
        return json.dumps({u'result': getResult(300, u'Превышен лимит привязок')}, ensure_ascii=False)

    if checkTerminal(request.query.get('cashCode'), request.query.get('shopCode')) == None:
        result = {u'result': getResult(300, u'Для данного терминала невозможно выполнить операцию')}
    elif len(bindingList[cardNum]) >= 5:
        result = {u'result': getResult(300, u'Превышен лимит привязок')}
        logging.error(u'Превышен лимит привязок')
    else:
        newAccount = {
                    "accountNumber":accNum,
                    "providerCode":provCode,
                    "providerName":providerList[provCode]["name"],
                    "modificationDate": datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f+0000'),
                    "status":"CHECKED"
                }
        bindingList[cardNum][accNum] = newAccount
        result = {u'result': getResult(0, u'OK')}
    logging.debug(u'Отправляем ответ: %s' % json.dumps(result, ensure_ascii=False))

    return json.dumps(result, ensure_ascii=False)

@route('/paygate3/api/bindings/cardnumber/<cardNum>/accountnumber/<accNum>', method='DELETE')
@auth_basic(checkPassword)
def unbindAccount(cardNum, accNum):
    cardNum = cardNum.decode('utf-8')
    logging.info(u'Получили запрос на отвязку номера')
    logging.info(u'URL: %s' % request.url)

    rule = getRule('unbindAccount')
    baseError = checkBaseRule(rule)
    if baseError:
        return json.dumps({u'result': getResult(300, baseError)}, ensure_ascii=False)
    if rule == 'accountNotFound':
        logging.error(u'Тест.Аккаунт не найден')
        return json.dumps({u'result': getResult(300, u'Аккаунт не найден')}, ensure_ascii=False)


    if checkTerminal(request.query.get('cashCode'), request.query.get('shopCode')) == None:
        result = {u'result': getResult(300, u'Для данного терминала невозможно выполнить операцию')}
    elif accNum not in bindingList[cardNum]:
        logging.error('Аккаунт не найден')
        result = {u'result': getResult(300, u'Аккаунт не найден')}
    else:
        bindingList[cardNum].pop(accNum)
        result = {u'result': getResult(0, u'OK')}
    logging.debug(u'Отправляем ответ: %s' % json.dumps(result, ensure_ascii=False))

    return json.dumps(result, ensure_ascii=False)

def init():
    logging.info(u'Чтение настроек')
    global restServiceHost
    global restServicePort
    global testMode
    global bindingList
    global providerList
    global terminalList
    global organizationList
    global paymentSystemList
    
    conf = ConfigParser.RawConfigParser()
    conf.read(servicePath + 'conf/config.ini')
    restServiceHost = conf.get('main', 'host')
    restServicePort = conf.getint('main', 'port')
    testMode = conf.getboolean('main', 'testMode')

    json_data=open(servicePath + 'data/binding.json').read()
    bindingList = json.loads(json_data)

    json_data=open(servicePath + 'data/provider.json').read()
    providerList = json.loads(json_data)

    json_data=open(servicePath + 'data/organization.json').read()
    organizationList = json.loads(json_data)

    json_data=open(servicePath + 'data/paymentsystem.json').read()
    paymentSystemList = json.loads(json_data)

    json_data=open(servicePath + 'data/terminal.json').read()
    terminalList = json.loads(json_data)

    if testMode:
        for key, val in conf.items('test'):
            rule = key.split('.')
            beg = 0
            end = 0
            func = rule[0]
            if not func in ruleList:
                ruleList[func] =[] 
            if len(rule) > 1:
                interval = rule[1].split('-')
                beg = int(interval[0])
                end = int(interval[0])
                if len(interval) > 1:
                    end = int(interval[1])
            i = beg
            while i <= end:
                while i >= len(ruleList[func]):
                    ruleList[func].append(None)
                ruleList[func][i] = val
                i +=1
        logging.info(u'Заданы правила для сервиса: %s' % ruleList)

class MySSLCherryPy(ServerAdapter):
    def run(self, handler):
        server = _cpwsgiserver3.CherryPyWSGIServer((self.host, self.port), handler)

        cert = servicePath + 'server.pem'
        server.ssl_certificate = cert
        server.ssl_private_key = cert
        try:
            server.start()
        finally:
            server.stop()

def start():
    logging.info(u'Запускаем сервис')
    try: 
        init()
        logging.info(u'Стартуем на %s:%s' % (restServiceHost, restServicePort))
        run(host=restServiceHost, port=restServicePort, server=MySSLCherryPy)
    except Exception, e:
        print str(e)

if __name__ == '__main__':
    start()
